from typing import Any

from pydantic import BaseModel, Field, computed_field, field_validator, model_validator

from akasha.constants import PERCENT_STAT_TYPES

from .. import enums

__all__ = ("Artifact", "ArtifactStat")


class ArtifactStat(BaseModel):
    type: enums.ArtifactStat = Field(alias="name")
    value: float

    @computed_field
    @property
    def display_value(self) -> str:
        if self.type in PERCENT_STAT_TYPES:
            return f"{self.value * 100:.1f}%"
        return str(round(self.value))


class Artifact(BaseModel):
    name: str
    set_name: str = Field(alias="setName")
    rarity: int = Field(alias="stars")
    main_stat: ArtifactStat
    substats: list[ArtifactStat]
    equip_type: enums.EquipType = Field(alias="equipType")
    level: int
    crit_value: float = Field(alias="critValue")
    icon: str

    @field_validator("substats", mode="before")
    @classmethod
    def __transform_substats(cls, v: dict[str, float]) -> list[ArtifactStat]:
        return [
            ArtifactStat(name=enums.ArtifactStat(name), value=value) for name, value in v.items()
        ]

    @field_validator("level", mode="before")
    @classmethod
    def __subtract_level(cls, v: int) -> int:
        return v - 1

    @model_validator(mode="before")
    @classmethod
    def __transform_main_stat(cls, v: dict[str, Any]) -> dict[str, Any]:
        v["main_stat"] = {"name": v.pop("mainStatKey"), "value": v.pop("mainStatValue")}
        return v



from enum import StrEnum

__all__ = ("ArtifactStat", "CharaStatType", "Element", "EquipType", "Language")


class Language(StrEnum):
    """Languages supported by Akasha System."""

    ENGLISH = "en"
    GERMAN = "de"
    FRENCH = "fr"
    SPANISH = "es"
    ITALIAN = "it"
    JAPANESE = "ja"
    KOREAN = "ko"
    PORTUGUESE = "pt"
    RUSSIAN = "ru"
    THAI = "th"
    TURKISH = "tr"
    VIETNAMESE = "vi"
    CHINESE_SIMPLIFIED = "zh-CN"
    CHINESE_TRADITIONAL = "zh-TW"


class EquipType(StrEnum):
    """Artifact types."""

    FLOWER = "EQUIP_BRACER"
    FEATHER = "EQUIP_NECKLACE"
    SANDS = "EQUIP_SHOES"
    GOBLET = "EQUIP_RING"
    CIRCLET = "EQUIP_DRESS"


class Element(StrEnum):
    """Elements in Genshin Impact."""

    HYDRO = "Hydro"
    ANEMO = "Anemo"
    PYRO = "Pyro"
    CRYO = "Cryo"
    ELECTRO = "Electro"
    GEO = "Geo"
    DENDRO = "Dendro"


class ArtifactStat(StrEnum):
    """Artifact stat names."""

    CRIT_DMG = "Crit DMG"
    CRIT_RATE = "Crit RATE"

    FLAT_HP = "Flat HP"
    HP_PERCENT = "HP%"
    FLAT_ATK = "Flat ATK"
    ATK_PERCENT = "ATK%"
    FLAT_DEF = "Flat DEF"
    DEF_PERCENT = "DEF%"

    ENERGY_RECHARGE = "Energy Recharge"
    ELEMENTAL_MASTERY = "Elemental Mastery"

    HYDRO_DMG_BONUS = "Hydro DMG Bonus"
    ANEMO_DMG_BONUS = "Anemo DMG Bonus"
    PYRO_DMG_BONUS = "Pyro DMG Bonus"
    CRYO_DMG_BONUS = "Cryo DMG Bonus"
    ELECTRO_DMG_BONUS = "Electro DMG Bonus"
    GEO_DMG_BONUS = "Geo DMG Bonus"
    DENDRO_DMG_BONUS = "Dendro DMG Bonus"


class CharaStatType(StrEnum):
    """Character stat types."""

    MAX_HP = "maxHp"
    ATK = "atk"
    DEF = "def"
    ELEMENTAL_MASTERY = "elementalMastery"
    ENERGY_RECHARGE = "energyRecharge"
    HEALING_BONUS = "healingBonus"
    CRIT_RATE = "critRate"
    CRIT_DMG = "critDamage"

    CRYO_DMG_BONUS = "cryoDamageBonus"
    PYRO_DMG_BONUS = "pyroDamageBonus"
    HYDRO_DMG_BONUS = "hydroDamageBonus"
    ANEMO_DMG_BONUS = "anemoDamageBonus"
    ELECTRO_DMG_BONUS = "electroDamageBonus"
    GEO_DMG_BONUS = "geoDamageBonus"
    DENDRO_DMG_BONUS = "dendroDamageBonus"



from .enums import ArtifactStat, CharaStatType

PERCENT_STAT_TYPES: set[CharaStatType | ArtifactStat] = {
    CharaStatType.CRIT_RATE,
    CharaStatType.CRIT_DMG,
    CharaStatType.CRYO_DMG_BONUS,
    CharaStatType.PYRO_DMG_BONUS,
    CharaStatType.HYDRO_DMG_BONUS,
    CharaStatType.ANEMO_DMG_BONUS,
    CharaStatType.ELECTRO_DMG_BONUS,
    CharaStatType.GEO_DMG_BONUS,
    CharaStatType.DENDRO_DMG_BONUS,
    CharaStatType.HEALING_BONUS,
    CharaStatType.ENERGY_RECHARGE,
    ArtifactStat.ENERGY_RECHARGE,
    ArtifactStat.CRIT_RATE,
    ArtifactStat.CRIT_DMG,
    ArtifactStat.HP_PERCENT,
    ArtifactStat.ATK_PERCENT,
    ArtifactStat.DEF_PERCENT,
    ArtifactStat.CRYO_DMG_BONUS,
    ArtifactStat.PYRO_DMG_BONUS,
    ArtifactStat.HYDRO_DMG_BONUS,
    ArtifactStat.ANEMO_DMG_BONUS,
    ArtifactStat.ELECTRO_DMG_BONUS,
    ArtifactStat.GEO_DMG_BONUS,
    ArtifactStat.DENDRO_DMG_BONUS,
}



from typing import Any, Literal

from pydantic import BaseModel, Field, computed_field, field_validator

__all__ = ("CharacterCalc", "CharacterCalcWeapon", "UserCalc", "UserCalcArtifact")


class CharacterCalcWeapon(BaseModel):
    name: str
    icon: str
    substat: str
    type: str
    rarity: Literal[1, 2, 3, 4, 5]
    refinement: Literal[1, 2, 3, 4, 5]


class CharacterCalc(BaseModel):
    id: int = Field(alias="calculationId")
    short: str
    name: str
    details: str
    weapon: CharacterCalcWeapon
    result: float
    stats: dict[str, float]
    ranking: int
    out_of: int = Field(alias="outOf")

    @computed_field
    @property
    def top_percent(self) -> float:
        return (self.ranking / self.out_of) * 100


class UserCalcArtifact(BaseModel):
    name: str
    icon: str
    count: int


class UserCalc(BaseModel):
    id: str = Field(alias="_id")
    character_id: int = Field(alias="characterId")
    type: str
    uid: str
    artifact_sets: list[UserCalcArtifact] = Field(alias="artifactSets")
    calculations: list[CharacterCalc]
    constellation: Literal[0, 1, 2, 3, 4, 5, 6]
    md5: str
    name: str
    icon: str

    @field_validator("artifact_sets", mode="before")
    @classmethod
    def __flatten_artifacts(cls, v: dict[str, Any]) -> list[UserCalcArtifact]:
        return [UserCalcArtifact(name=name, **artifact) for name, artifact in v.items()]

    @field_validator("calculations", mode="before")
    @classmethod
    def __flatten_calculations(cls, v: dict[str, Any]) -> list[CharacterCalc]:
        return [CharacterCalc(**calc) for calc in v.values()]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Self

from pydantic import BaseModel, Field, computed_field, field_validator, model_validator

from akasha.constants import PERCENT_STAT_TYPES
from akasha.enums import CharaStatType, Element

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

__all__ = ("Leaderboard", "LeaderboardCalc", "LeaderboardOwner", "ProfilePicture")


class LeaderboardCalc(BaseModel):
    id: str
    result: float


class ProfilePicture(BaseModel):
    icon: str | None = None
    old_icon: str | None = Field(None, alias="oldIcon")

    @field_validator("icon", "old_icon", mode="after")
    @classmethod
    def __convert_icon(cls, v: str) -> str:
        return f"https://enka.network/ui/{v}.png"

    @model_validator(mode="before")
    @classmethod
    def __flatten_assets(cls, v: dict[str, Any]) -> dict[str, Any]:
        if not isinstance(v, dict):
            return {}
        assets = v.pop("assets", {})
        return {**v, **assets}


class LeaderboardOwner(BaseModel):
    nickname: str
    adventure_rank: int = Field(alias="adventureRank")
    profile_picture: ProfilePicture = Field(alias="profilePicture")
    namecard: str = Field(alias="nameCard")
    region: str

    @field_validator("adventure_rank", mode="before")
    @classmethod
    def __intify_adventure_rank(cls, v: float) -> int:
        return int(v)

    @field_validator("namecard", mode="after")
    @classmethod
    def __convert_namecard(cls, v: str) -> str:
        return f"https://enka.network/ui/{v}.png"


class CharacterTalent(BaseModel):
    icon: str
    level: int
    raw_level: int = Field(alias="rawLevel")
    id: int
    boosted: bool


class CharacterWeapon(BaseModel):
    level: int
    ascension: int = Field(alias="promoteLevel")
    refinement: int
    icon: str
    name: str

    @model_validator(mode="before")
    @classmethod
    def __unnest_weapon_info(cls, v: dict[str, Any]) -> dict[str, Any]:
        weapon_info = v.pop("weaponInfo")
        v["level"] = weapon_info["level"]
        v["promoteLevel"] = weapon_info["promoteLevel"]
        v["refinement"] = weapon_info["refinementLevel"]["value"]
        return v


class CharacterStat(BaseModel):
    type: CharaStatType
    value: float

    @computed_field
    @property
    def display_value(self) -> str:
        if self.type in PERCENT_STAT_TYPES:
            return f"{self.value * 100:.1f}%"
        return str(round(self.value))


class Leaderboard(BaseModel):
    id: str = Field(alias="_id")
    calculation: LeaderboardCalc
    build_name: str = Field(alias="type")
    uid: str
    owner: LeaderboardOwner
    rank: int = Field(alias="index")
    md5: str

    # Character related fields
    character_id: int = Field(alias="characterId")
    constellation: int
    crit_value: float = Field(alias="critValue")
    ascension: int
    level: int
    stats: dict[CharaStatType, CharacterStat]
    talents: dict[str, CharacterTalent] = Field(alias="talentsLevelMap")
    weapon: CharacterWeapon
    element: Element = Field(alias="characterMetadata")

    @field_validator("stats", mode="before")
    @classmethod
    def __unnest_stats_value(
        cls, v: dict[str, dict[str, float]]
    ) -> dict[CharaStatType, CharacterStat]:
        return {
            CharaStatType(k): CharacterStat(type=CharaStatType(k), value=v["value"])
            for k, v in v.items()
        }

    @field_validator("element", mode="before")
    @classmethod
    def __unnest_element(cls, v: dict[str, str]) -> str:
        return v["element"]

    @model_validator(mode="before")
    @classmethod
    def __unnest_fields(cls, v: dict[str, Any]) -> dict[str, Any]:
        prop_map = v.pop("propMap")
        v["ascension"] = int(prop_map["ascension"]["val"] or 0)
        v["level"] = int(prop_map["level"]["val"] or 1)
        return v


class LeaderboardPaginator:
    def __init__(
        self,
        fetch_boards: Callable[[int, int, int, str, bool], Awaitable[list[Leaderboard]]],
        calculation_id: int,
        page_size: int,
        max_page: int,
        use_cache: bool,
    ) -> None:
        self._fetch_boards = fetch_boards
        self._boards: list[Leaderboard] = []

        self._calculation_id = calculation_id
        self._page = 1
        self._index = 0
        self._page_size = page_size
        self._max_page = max_page
        self._use_cache = use_cache

    def __aiter__(self) -> Self:
        return self

    async def __anext__(self) -> Leaderboard:
        if self._max_page < 0:
            msg = "max_page must be greater than 0"
            raise ValueError(msg)

        if self._index >= len(self._boards) - 1:
            if self._page > self._max_page:
                raise StopAsyncIteration
            self._boards = await self._fetch_boards(
                self._calculation_id,
                self._page,
                self._page_size,
                f"lt|{self._boards[-1].calculation.result}" if self._boards else "",
                self._use_cache,
            )
            self._index = 0
            self._page += 1
        else:
            self._index += 1

        return self._boards[self._index]



from typing import Any, Final, Self

from aiohttp_client_cache.backends.sqlite import SQLiteBackend
from aiohttp_client_cache.session import CachedSession
from loguru import logger

from akasha.models.leaderboard import LeaderboardPaginator

from .enums import Language
from .errors import DESC_TO_ERROR, AkashaAPIError
from .models import Leaderboard, UserCalc
from .models.artifact import Artifact

__all__ = ("AkashaAPI",)


class AkashaAPI:
    BASE_URL: Final[str] = "https://akasha.cv/api"

    def __init__(
        self,
        lang: Language = Language.ENGLISH,
        headers: dict[str, Any] | None = None,
        cache_name: str = "./.cache/akasha-py.db",
        cache_ttl: int = 360,
        debug: bool = False,
    ) -> None:
        self._lang = lang
        self._headers = headers or {"User-Agent": "akasha-py"}
        self._cache_name = cache_name
        self._cache_ttl = cache_ttl
        self._session: CachedSession | None = None
        self.debug = debug

    async def __aenter__(self) -> Self:
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:  # noqa: ANN001
        await self.close()

    async def start(self) -> None:
        cache = SQLiteBackend(self._cache_name, expire_after=self._cache_ttl)
        self._session = CachedSession(cache=cache)

    async def close(self) -> None:
        if self._session is None:
            return
        await self._session.close()

    def _raise_for_error(self, data: list[dict[str, Any]] | dict[str, Any]) -> None:
        if isinstance(data, list):
            return

        if (error := data.get("error")) is not None:
            if (message := data.get("message")) is not None:
                title = message
                description = error
            else:
                title = error["title"]
                description = error["description"]

            if (error_class := DESC_TO_ERROR.get(description)) is not None:
                raise error_class(title, description)
            raise AkashaAPIError(title, description)

    async def _request(
        self, endpoint: str, use_cache: bool, *, params: dict[str, Any] | None = None
    ) -> Any:
        if self._session is None:
            msg = f"Session is not started, call {self.__class__.__name__}.start() first."
            raise RuntimeError(msg)

        url = f"{self.BASE_URL}/{endpoint}"
        params = params or {}
        if self.debug:
            logger.debug(f"Requesting {url} with params {params}")

        if not use_cache:
            async with self._session.disabled(), self._session.get(url, params=params) as response:
                response.raise_for_status()
                data = await response.json()
        else:
            async with self._session.get(url, params=params) as response:
                response.raise_for_status()
                data = await response.json()

        data = data.get("data", data)
        self._raise_for_error(data)
        return data

    async def get_calculations_for_user(
        self, uid: int, *, use_cache: bool = True
    ) -> list[UserCalc]:
        """Get the calculations for a user.

        Args:
            uid: The user ID.
            use_cache: Whether to use the cache.
        """
        data = await self._request(f"getCalculationsForUser/{uid}", use_cache=use_cache)
        user_calcs = [UserCalc(**calc) for calc in data]

        strings: set[str] = set()
        for user_calc in user_calcs:
            strings.add(user_calc.name)
            for calc in user_calc.calculations:
                strings.add(calc.weapon.name)

        translations = await self.get_translations(list(strings))

        for user_calc in user_calcs:
            user_calc.name = translations.get(user_calc.name.lower(), user_calc.name)
            for calc in user_calc.calculations:
                calc.weapon.name = translations.get(calc.weapon.name.lower(), calc.weapon.name)

        return user_calcs

    async def _fetch_leaderboards(
        self, calculation_id: int, page: int, page_size: int, p: str, use_cache: bool
    ) -> list[Leaderboard]:
        data = await self._request(
            "leaderboards",
            params={
                "calculationId": calculation_id,
                "size": page_size,
                "page": page,
                "sort": "calculation.result",
                "order": -1,
                "p": p,
            },
            use_cache=use_cache,
        )
        return [Leaderboard(**lb) for lb in data]

    def get_leaderboards(
        self, calculation_id: int, max_page: int, *, page_size: int = 20, use_cache: bool = True
    ) -> LeaderboardPaginator:
        """Get a leaderboard paginator for a calculation.

        Args:
            calculation_id: The calculation ID.
            max_page: The maximum number of pages to return.
            page_size: The number of leaderboards to return per page.
            use_cache: Whether to use the cache.
        """
        return LeaderboardPaginator(
            self._fetch_leaderboards, calculation_id, page_size, max_page, use_cache
        )

    async def refresh_user(self, uid: int) -> None:
        """Refresh the Enka data of a player.

        Args:
            uid: The UID of the player.
        """
        await self._request(f"user/refresh/{uid}", use_cache=False)

    async def get_user(self, uid: int, *, use_cache: bool = True) -> None:
        """Get the user data.

        Args:
            uid: The UID of the player.
            use_cache: Whether to use the cache.
        """
        msg = "This method is not implemented yet."
        raise NotImplementedError(msg)
        await self._request(f"user/{uid}", use_cache=use_cache)

    async def get_translations(self, words: list[str], *, use_cache: bool = True) -> dict[str, str]:
        """Get the translations for a list of words.

        Args:
            words: The list of words to translate.
            use_cache: Whether to use the cache.

        Returns:
            A dictionary of the translations. Key is word.lower(), value is the translation.
        """
        if self._lang is Language.ENGLISH:
            return {word.lower(): word for word in words}

        data = await self._request(
            f"textmap/{self._lang.value}", use_cache=use_cache, params={"words[]": words}
        )
        return data["translation"]

    async def translate(self, word: str) -> str:
        """Translate a word.

        Do not use this method for multiple translations, use get_translations instead.

        Args:
            word: The word to translate.

        Returns:
            The translation of the word.
        """
        if self._lang is Language.ENGLISH:
            return word

        return (await self.get_translations([word]))[word.lower()]

    async def get_artifacts(self, uid: int, md5: str, *, use_cache: bool = True) -> list[Artifact]:
        """Get the artifacts of a build.

        Args:
            uid: The UID of the owner of the build.
            md5: The MD5 hash of the build.
            use_cache: Whether to use the cache.
        """
        data = await self._request(f"artifacts/{uid}/{md5}", use_cache=use_cache)
        return [Artifact(**artifact) for artifact in data]



__all__ = ("DESC_TO_ERROR", "AkashaAPIError", "UIDNotFoundError")


class AkashaAPIError(Exception):
    """Base class for all Akasha API errors."""

    def __init__(self, title: str, description: str) -> None:
        self.title = title
        self.description = description

    def __str__(self) -> str:
        return f"{self.title}: {self.description}"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.title!r}, {self.description!r})"


class UIDNotFoundError(AkashaAPIError):
    """Raised when the provided UID cannot be fetched from Enka Network API."""


class IncorrectUIDFormatError(AkashaAPIError):
    """Raised when the provided UID is not in the correct format."""


DESC_TO_ERROR = {
    "Provided UID cannot be fetched from Enka Network API": UIDNotFoundError,
    "The UID format is incorrect": IncorrectUIDFormatError,
}
